<cfcomponent output="false">

	<cfset System = createObject("java", "java.lang.System") />
	<cfset LIB_DIRECTORY = getDirectoryFromPath(getCurrentTemplatePath()) & "lib/" />
	
	<cfset RHINO_OBJECT_CLASS = "org.mozilla.javascript.NativeObject" />
	<cfset RHINO_ARRAY_CLASS = "org.mozilla.javascript.NativeArray" />
	<cfset RHINO_DATE_CLASS = "org.mozilla.javascript.NativeDate" />
	<cfset RHINO_UNDEFINED_CLASS = "org.mozilla.javascript.Undefined" />

	<cfset RHINO_JAVA_OBJECT_CLASS = "org.mozilla.javascript.NativeJavaObject" />

	<cffunction name="init" access="public" output="false" returntype="rhino">
		<cfargument name="scriptPaths" type="any" default="#arrayNew(1)#" />
		<cfif isSimpleValue(scriptPaths)>
			<cfset variables.my.scriptPaths = arrayNew(1) />
			<cfset arrayAppend(my.scriptPaths, scriptPaths) />
		<cfelse>
			<cfset variables.my.scriptPaths = scriptPaths />
		</cfif>
		<cfset arrayAppend(my.scriptPaths, LIB_DIRECTORY) />
		<cfset variables.my.scriptCache = structNew() />
		<cfreturn this />
	</cffunction>
	
	
	
	<cffunction name="runScript" access="public" output="false" returntype="struct">
		<cfargument name="filename" type="string" required="true" />
		<cfargument name="params" type="struct" default="#structNew()#" />
		<cfset var result = "" />
		<cfset var bindings = structNew() />
		<cfset var internal = structNew() />
		<cfset var Context = createObject("java", "org.mozilla.javascript.Context") />
		<cfset var ctx = "" />
		<cfset var scope = "" />
		<cfset var i = "" />
		<cfset var totalStart = getTickCount() />
		<cfset var oFile = "" />
		<cfset var bootstrapFilename = "" />
		<cfset var reader = "" />
		<cfset var script = "" />
		
		<!--- set up the bindings for the script execution --->
		<cfset bindings["params"] = convertToRhino(params) />
		<cfset bindings["model"] = convertToRhino(structNew()) />
		<cfset bindings["__internal__"] = internal /> <!--- to be populated below --->

		<cfset internal["scriptCache"] = my.scriptCache />

		<cftry>
			<cfset ctx = Context.enter() />
			<cfset ctx.setApplicationClassLoader(getClass().getClassLoader()) />
			
			<!--- ensure the script has been compiled --->
			<!--- this caching must match that in core/bootstrap.js's loadScript method --->
			<cfset oFile = createObject("java", "java.io.File").init(LIB_DIRECTORY & "core/bootstrap.js") />
			<cfset bootstrapFilename = oFile.getCanonicalPath() />
			<cfif NOT structKeyExists(internal.scriptCache, bootstrapFilename)
					OR internal.scriptCache[bootstrapFilename].timestamp.getTime() LT oFile.lastModified()>
				<!--- if the bootstrapper changes, we need to recompile everything --->
				<cfset structClear(internal.scriptCache) />
				
				<!--- set up the reader and compile --->
				<cfset reader = createObject("java", "java.io.BufferedReader").init(createObject("java", "java.io.FileReader").init(oFile)) />
				<cfset script = ctx.compileReader(
					reader,
					bootstrapFilename,
					1,
					javaCast("null", 0)
				) />
				<cfset reader.close() />
				
				<!--- cache the compiled script --->
				<cfset internal.scriptCache[bootstrapFilename] = {
					timestamp = now(),
					script = script
				} />
			</cfif>
			
			<cfset scope = ctx.initStandardObjects() />

			<cfset script = createObject("component", "rhinoRequest").init(
				internal.scriptCache[oFile.getCanonicalPath()].script
			) />
			<!--- set up the internal data structures --->
			<cfset internal["context"] = ctx />
			<cfset internal["global_scope"] = scope />
			<cfset internal["pageContext"] = script.getLocalPageContext() />
			<cfset internal["request"] = internal["pageContext"].getRequest() />
			<cfset internal["response"] = internal["pageContext"].getResponse() />
			<cfset internal["searchPath"] = convertToRhino(my.scriptPaths) />
			<cfset internal["scriptToRun"] = filename />

			<!--- load the bindings into the global scope --->
			<cfloop collection="#bindings#" item="i">
				<cfset scope.putProperty(scope, i, bindings[i]) />
			</cfloop>
			
			<!--- run the script! --->
			<cfset script.execute(ctx, scope) />
			<cfset Context.exit() />
			
			<cfcatch type="any">
				<cfset Context.exit() />
				<cftry>
					<cfset reader.close() />
					<cfcatch type="any"><!--- no worries ---></cfcatch>
				</cftry>
				<cfif cfcatch.type EQ "org.mozilla.javascript.JavaScriptException" AND cfcatch.message CONTAINS "Early Exit:">
					<!--- ignore it --->
				<cfelseif structKeyExists(cfcatch, "wrappedException")>
					<cfthrow object="#cfcatch.wrappedException#" />
				<cfelse>
					<cfrethrow />
				</cfif>
			</cfcatch>
		</cftry>

		<cfset result = restoreFromRhino(bindings["model"]) />
		<cfset result["RHINO_TOTAL_EXECUTION_TIME"] = getTickCount() - totalStart />

		<cfreturn result />
	</cffunction>
	
	
	
	<cffunction name="convertToRhino" access="private" output="false" returntype="any">
		<cfargument name="data" type="any" required="true" />
		<cfargument name="visited" type="any" default="" />
		<cfset var result = "" />
		<cfset var i = "" />
		<cfset var hashCode = System.identityHashCode(data) />
		<cfset var row = "" />
		<cfif isSimpleValue(visited)>
			<cfset visited  = createObject("java", "java.util.HashMap").init() />
		</cfif>
		<cfif isSimpleValue(data)>
			<cfreturn data />
		<cfelseif visited.containsKey(hashCode)>
			<cfreturn visited.get(hashCode) />
		<cfelse>
			<cfif isStruct(data)>
				<cfset result = createObject("java", RHINO_OBJECT_CLASS).init() />
				<cfset visited.put(hashCode, result) />
				<cfloop collection="#data#" item="i">
					<cfset result.putProperty(result, i, convertToRhino(data[i], visited)) />
				</cfloop>
			<cfelseif isArray(data)>
				<cfset result = createObject("java", RHINO_ARRAY_CLASS).init(javaCast("int", arrayLen(data))) />
				<cfset visited.put(hashCode, result) />
				<cfloop from="1" to="#arrayLen(data)#" index="i">
					<cfset result.putProperty(result, javaCast("int", i - 1), convertToRhino(data[i], visited)) />
				</cfloop>
			<cfelseif isQuery(data)>
				<cfset result = createObject("java", RHINO_ARRAY_CLASS).init(javaCast("int", data.recordCount)) />
				<cfset visited.put(hashCode, result) />
				<cfloop query="data">
					<cfset row = createObject("java", RHINO_OBJECT_CLASS).init() />
					<cfloop list="#data.columnList#" index="i">
						<cfset row.putProperty(result, i, convertToRhino(data[i][currentRow], visited)) />
					</cfloop>
					<cfset result.putProperty(result, javaCast("int", currentRow - 1), row) />
				</cfloop>
			<cfelseif isObject(data)>
				<cfset result = data />
			<cfelse>
				<cfthrow type="IllegalArgumentException"
					message="The passed object is not of a recognized type.  It's a '#data.getClass().getName()#'." />
			</cfif>
			<cfreturn result />
		</cfif>
	</cffunction>
	
	
	
	<cffunction name="restoreFromRhino" access="public" output="false" returntype="any">
		<cfargument name="data" type="any" required="false" />
		<cfargument name="visited" type="any" default="" />
		<cfset var result = "" />
		<cfset var ids = "" />
		<cfset var i = "" />
		<cfset var hashCode = "" />
		<cfif NOT structKeyExists(arguments, "data")>
			<cfreturn />
		</cfif>
		<cfif isSimpleValue(visited)>
			<cfset visited  = createObject("java", "java.util.HashMap").init() />
		</cfif>
		<cfset hashCode = System.identityHashCode(data) />
		<cfif isSimpleValue(data)>
			<cfreturn data />
		<cfelseif visited.containsKey(hashCode)>
			<cfreturn visited.get(hashCode) />
		<cfelse>
			<cfif data.getClass().getName() EQ RHINO_OBJECT_CLASS>
				<cfset result = structNew() />
				<cfset visited.put(hashCode, result) />
				<cfset ids = createObject("java", "java.util.Arrays").asList(data.getIds()) />
				<cfloop from="1" to="#arrayLen(ids)#" index="i">
					<cfset result[ids[i]] = restoreFromRhino(data.getProperty(data, ids[i]), visited) />
				</cfloop>
			<cfelseif data.getClass().getName() EQ RHINO_ARRAY_CLASS>
				<cfset result = arrayNew(1) />
				<cfset visited.put(hashCode, result) />
				<cfset ids = createObject("java", "java.util.Arrays").asList(data.getIds()) />
				<cfloop from="1" to="#arrayLen(ids)#" index="i">
					<cfif isNumeric(ids[i])>
						<cfset result[ids[i] + 1] = restoreFromRhino(data.getProperty(data, ids[i]), visited) />
					</cfif>
				</cfloop>
			<cfelseif data.getClass().getName() EQ RHINO_DATE_CLASS>
				<cfset result = createObject("java", "java.util.Date").init(data.callMethod(data, "getTime", javaCast("null", 0))) />
				<cfset visited.put(hashCode, result) />
			<cfelseif data.getClass().getName() EQ RHINO_UNDEFINED_CLASS>
				<cfset result = "undefined" />
			<cfelseif data.getClass().getName() EQ RHINO_JAVA_OBJECT_CLASS>
				<cfset result = data.unwrap() />
			<cfelse>
				<cfset result = data />
			</cfif>
			<cfreturn result />
		</cfif>
	</cffunction>

<cfoutput></cfoutput>
</cfcomponent>